import unittest

from tests import utils

from pipecash.pipeScheduler import Scheduler


class PipeSchedulerTest(unittest.TestCase):

    def setUp(self):
        self.schedulerInstance = Scheduler()

    def tearDown(self):
        self.schedulerInstance.stop()

    def test_scheduler_knownSchedules(self):
        knownSchedules_every = [
            "every_1s", "every_2s", "every_5s", "every_10s", "every_30s",
            "every_1m", "every_2m", "every_5m", "every_10m", "every_30m",
            "every_1h", "every_2h", "every_5h", "every_12h",
            "every_1d", "every_2d", "every_7d", "every_30d"
        ]
        knownSchedules_at = [
            "midnight", "1am", "2am", "3am", "4am", "5am",
            "6am", "7am", "8am", "9am", "10am", "11am",
            "noon", "1pm", "2pm", "3pm", "4pm", "5pm",
            "6pm", "7pm", "8pm", "9pm", "10pm", "11pm",
        ]

        for sch in knownSchedules_every + knownSchedules_at:
            self.assertTrue(self.schedulerInstance.isKnownSchedule(sch), "")

        unKnownSchedules = [
            "afternoon", "1:30am", "01:00", "13:00",
            "every 30s", "every_42s", "every_30ms"
            "batman", "", 0, 42, None, {}
        ]

        for sch in unKnownSchedules:
            self.assertFalse(self.schedulerInstance.isKnownSchedule(sch), "")

    def test_scheduler_knownSchedules(self):
        knownSchedules_every = [
            "every_1s", "every_2s", "every_5s", "every_10s", "every_30s",
            "every_1m", "every_2m", "every_5m", "every_10m", "every_30m",
            "every_1h", "every_2h", "every_5h", "every_12h",
            "every_1d", "every_2d", "every_7d", "every_30d"
        ]
        knownSchedules_at = [
            "midnight", "1am", "2am", "3am", "4am", "5am",
            "6am", "7am", "8am", "9am", "10am", "11am",
            "noon", "1pm", "2pm", "3pm", "4pm", "5pm",
            "6pm", "7pm", "8pm", "9pm", "10pm", "11pm",
        ]

        for sch in knownSchedules_every + knownSchedules_at:
            self.assertTrue(self.schedulerInstance.isKnownSchedule(sch), "")

        unKnownSchedules = [
            "afternoon", "1:30am", "01:00", "13:00",
            "every 30s", "every_42s", "every_30ms"
            "batman", "", 0, 42, None, {}
        ]
        for sch in unKnownSchedules:
            self.assertFalse(self.schedulerInstance.isKnownSchedule(sch), "")

    def test_scheduler_nameToSeconds(self):
        secondsPerLetter = {"s": 1, "m": 60, "h": 60*60, "d": 60*60*24}

        letter = 's'
        for letter in ['s', 'm', 'h', 'd']:
            for i in range(0, 60):
                self.assertEqual(
                    self.schedulerInstance.everyScheduleToSeconds(
                        "every_%s%s" % (i, letter)),
                    i * secondsPerLetter[letter])

    def test_scheduler_scheduleSimpleTask(self):
        varName = "test_scheduler_scheduleSimpleTask_1s"

        self.schedulerInstance.registerTask(
            "every_1s",
            lambda: setattr(self, varName, 42))

        self.schedulerInstance.start()

        t = utils.waitLoop(
            "wait for 1s schedule to set the variable",
            2, lambda: hasattr(self, varName))

        var = getattr(self, varName)

        self.assertGreater(t, 0.95)
        self.assertLess(t, 1.05)
        self.assertEqual(var, 42)

    def test_scheduler_scheduleSimpleTask2s(self):
        varName = "test_scheduler_scheduleSimpleTask_2s"
        self.schedulerInstance.registerTask(
            "every_2s",
            lambda: setattr(self, varName, 42))

        self.schedulerInstance.start()

        t = utils.waitLoop(
            "wait for 2s schedule to set the variable",
            3, lambda: hasattr(self, varName))

        var = getattr(self, varName)

        self.assertGreater(t, 1.95)
        self.assertLess(t, 2.05)
        self.assertEqual(var, 42)

    def test_scheduler_scheduleMultipleSimpleTasks(self):
        varName = "test_scheduler_scheduleMultipleSimpleTasks_var"

        for i in [1, 2, 3]:
            setattr(self, varName + str(i), 0)

        self.schedulerInstance.registerTask("every_1s",
                                            lambda: setattr(self, varName + "1", getattr(self, varName + "1") + 1))
        self.schedulerInstance.registerTask("every_1s",
                                            lambda: setattr(self, varName + "2", getattr(self, varName + "2") + 2))
        self.schedulerInstance.registerTask("every_1s",
                                            lambda: setattr(self, varName + "3", getattr(self, varName + "3") + 3))

        # Over 5 seconds, the variables will look like:
        #           1s    2s    3s    4s    5s
        # var 1 =   1,    2,    3,    4,    5
        # var 2 =   2,    4,    6,    8,    10
        # var 3 =   3,    6,    9,    12,   15

        self.schedulerInstance.start()

        t = utils.waitLoop("measure time to set all variables", 7,
                           lambda: getattr(self, varName + "1") == 5 and
                           getattr(self, varName + "2") == 10 and
                           getattr(self, varName + "3") == 15)

        self.assertGreater(t, 4.80, "Time: " + str(t))
        self.assertLess(t, 5.5, "Time: " + str(t))

        variables = [
            getattr(self, varName + "1"),
            getattr(self, varName + "2"),
            getattr(self, varName + "3")
        ]

        for i in [1, 2, 3]:
            self.assertEqual(variables[i-1], 5*i,
                             "Variables: " + str(variables))



import unittest
import time

from pipecash.agentWrapper import AgentWrapper
from pipecash import logWrapper

import tests.agent_test.mockAgents as mockAgents
from tests.logMock import LogMock

logMock = LogMock(logWrapper.loggerInstance)


class AgentReceiveTest(unittest.TestCase):

    def test_agentReceiveEvent(self):
        sender = mockAgents.EventCreator_MockAgent()
        target = mockAgents.EventReceiver_MockAgent()

        senderWrap = AgentWrapper(sender, {"name": "MockAgent1"}, {})

        targetWrap = AgentWrapper(target, {"name": "MockAgent2"}, {})

        targetWrap.receiveEventsFrom(senderWrap)

        senderWrap.start()
        targetWrap.start()

        with logMock:
            senderWrap._AgentWrapper__runCheck()
            time.sleep(0.005)

        expectedSenderMethods = [
            'validate_options', 'check_dependencies_missing', 'start', 'check'
        ]
        expectedTargetMethods = [
            'validate_options', 'check_dependencies_missing', 'start', 'receive'
        ]

        self.assertListEqual(sender.calledMethods, expectedSenderMethods,
                             "Checking called methods of sender...")
        self.assertListEqual(target.calledMethods, expectedTargetMethods,
                             "Checking called methods of target...")
        self.assertListEqual(target.receivedData, [42])
        self.assertListEqual(logMock.logs, [
            "Running [Check] on EventCreator_MockAgent",
            "Running [Receive] on EventReceiver_MockAgent" ])



import time
import sys
import traceback


def expectError(callable, message):
    try:
        callable()
    except Exception as ex:
        tb = traceback.format_exc()
        return ex, tb
    raise AssertionError(
        "Action was successful, but should have failed : %s\n" % message)


def raiseErr(error):
    raise error


def funcToMethod(func):
    class FuncHolder:
        def transformed(*args):
            return func()
    return FuncHolder().transformed


def methodMock(callCollectionObject, func, functionName):
    if not hasattr(callCollectionObject, "calledMethods"):
        callCollectionObject.calledMethods = []

    def mockWrapper(*args):
        callCollectionObject.calledMethods.append(functionName)
        return func(*args)
    return funcToMethod(mockWrapper)


def waitLoop(loop_message, time_s, conditionFunc):
    start = time.time()
    result = False
    while(result != True):
        t = time.time()
        if(t - start > time_s):
            raise TimeOutError(loop_message, time_s)
        result = conditionFunc()
    return time.time() - start


def __try(func):
    def wrapper():
        try:
            func()
            return True
        except Exception as ex:
            return False
    return wrapper


def tryWaitLoop(loop_message, time_s, tryFunc):
    return waitLoop(loop_message, time_s, __try(tryFunc))


class TimeOutError(Exception):
    def __init__(self, message, time_s):
        self.message = "Timed Out iterating on : '%s' after %s s" % (
            message, time_s)
        super(TimeOutError, self).__init__(self.message)



from __future__ import absolute_import, print_function

import json

from pipecash import validationUtils as validate

from pipecash.observedMethod import observedMethodInstance as obs


class SecretsManager:
    __secrets = {}

    def __init__(self, values):
        self.__secrets = values

    @staticmethod
    def loadFromFile(path):
        with open(path, "r") as read_file:
            secretsValues = json.load(read_file)
            validate.objectType(secretsValues, dict,
                                "The contents of '%s'" % path)
            return SecretsManager(secretsValues)

    @obs.observedMethod
    def hasKey(self, key):
        return key in self.__secrets

    @obs.observedMethod
    def findMissingKeys(self, *keys):
        missingKeys = [k for k in keys if k not in self.__secrets]
        return missingKeys

    @obs.observedMethod
    def get(self, *keys):
        missingKeys = self.findMissingKeys(*keys)
        if len(missingKeys) > 0:
            raise KeyError("Not all keys were found : " + repr(missingKeys))
        return dict((k, self.__secrets[k]) for k in keys)

    @obs.observedMethod
    def getAll(self):
        return self.__secrets



import unittest

from pipecash.scenario import Scenario

from pipecash import pipeScheduler
from pipecash import pipeObserver


class ScenarioTest(unittest.TestCase):

    scenarioPath = "./tests/scenario_test/test_scenario.json"

    def setUp(self):
        pipeScheduler.schedulerInstance = pipeScheduler.Scheduler()
        pipeObserver.observerInstance = pipeObserver.Observer()

    def test_loadScenario(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

    def test_loadScenario_checkNames(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        agentNames = [
            'Agent 0 - has wallet',
            'Agent 1 - receive',
            'Agent 2 - receive',
            'Agent 3 - controled'
        ]
        self.assertListEqual([i.name for i in scenario.agents], agentNames)

        walletNames = ['Test Wallet 0', 'Test Wallet 1 - unused']
        self.assertListEqual([i.name for i in scenario.wallets], walletNames)

    def test_loadScenario_walletsOnlyWhereConfigured(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        wallets = [scenario.wallets[0].wallet, None, None, None]
        self.assertListEqual(
            [i.agent.wallet for i in scenario.agents], wallets)
        walletWraps = [scenario.wallets[0], None, None, None]
        self.assertListEqual(
            [i.walletWrapper for i in scenario.agents], walletWraps)

    def test_loadScenario_calledMethods(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        calledAgentMethods = [
            ['validate_options', 'check_dependencies_missing', 'start'],
            ['validate_options', 'check_dependencies_missing', 'start'],
            ['validate_options', 'check_dependencies_missing', 'start'],
            ['validate_options', 'check_dependencies_missing', 'start'],
        ]
        for i in range(len(scenario.agents)):
            self.assertListEqual(
                scenario.agents[i].agent.calledMethods, calledAgentMethods[i])
        calledWalletMethods = [
            ['validate_options', 'check_dependencies_missing', 'start'],
            ['validate_options', 'check_dependencies_missing', 'start'],
        ]
        for i in range(len(scenario.wallets)):
            self.assertListEqual(
                scenario.wallets[i].wallet.calledMethods, calledWalletMethods[i])

    def test_loadScenario_options(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        agentOptions = [{"foo": "bar"}, {"foo": "bar"},
                        {"foo": "default_value"}, {}]
        self.assertListEqual(
            [i.agent.options for i in scenario.agents], agentOptions)

        walletOptions = [{}, {"foo": "default_value"}]
        self.assertListEqual(
            [i.wallet.options for i in scenario.wallets], walletOptions)

    def test_loadScenario_schedule(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        scheduledTasks = pipeScheduler.schedulerInstance._Scheduler__scheduledTasks
        keysWithTasks = [
            s for s in scheduledTasks if len(scheduledTasks[s]) > 0]

        self.assertListEqual(sorted(keysWithTasks),
                             sorted(["every_1s", "every_2s", "every_5s", "every_10s"]))

        self.assertListEqual([len(scheduledTasks[k])
                              for k in keysWithTasks], [1, 1, 1, 1])

    def test_loadScenario_eventsAndControl(self):
        scenario = Scenario(self.scenarioPath)
        scenario.prepareToStart({})
        scenario.start()

        id0 = scenario.agents[0]._AgentWrapper__id
        id1 = scenario.agents[1]._AgentWrapper__id
        id2 = scenario.agents[2]._AgentWrapper__id

        listeners = pipeObserver.observerInstance._Observer__listeners

        self.assertEqual(len(listeners), 3)

        listeners0 = listeners[str((None, 'Event', id0))]
        listeners1 = listeners[str((None, 'Event', id1))]
        listeners2 = listeners[str((None, 'Event', id2))]

        self.assertEqual(len(listeners0), 2)
        self.assertEqual(len(listeners1), 1)
        self.assertEqual(len(listeners2), 1)

        self.assertListEqual(listeners0, [
            scenario.agents[1]._AgentWrapper__receiveEvent,
            scenario.agents[2]._AgentWrapper__receiveEvent
        ])

        self.assertListEqual(
            listeners1, [scenario.agents[3]._AgentWrapper__receiveControlEvent])
        self.assertListEqual(
            listeners2, [scenario.agents[3]._AgentWrapper__receiveControlEvent])



import unittest
import time

from pipecash.observedMethod import ObservedMethod
from pipecash.pipeObserver import Observer

ArgListNoArgs = str({'args': (), 'kwargs': {}})
ArgListWithArgs = str({'args': (1, 2), 'kwargs': {}})

ResultNone = "{'result': None}"
ResultOne = "{'result': 1}"

ExceptoionMsg = "{'error': AssertionError('testing the error',)}"

lines = []
linesFormat = "%s from %s with data: %s"


def __logObservations(area, name, state, sender, data):
    line = linesFormat % (str([area, name, state]), sender.__name__, str(data))
    lines.append(line)


observer = Observer()
obs = ObservedMethod()
obs.observerInstance = observer
observer.listen(None, None, None, __logObservations)


class ObservedMethodTest(unittest.TestCase):

    def __getLine(self, area, name, state, data):
        return linesFormat % (
            repr([area, name, state]), name, data)

    def test_observedFunction_successfulCall(self):
        lines[:] = []
        success()
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine(None, "success", "Start", ArgListNoArgs),
            self.__getLine(None, "success", "Done", ResultNone)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedFunction_withArgs_successfulCall(self):
        lines[:] = []
        success(1, 2)
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine(None, "success", "Start", ArgListWithArgs),
            self.__getLine(None, "success", "Done", ResultOne)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedFunction_failedCall(self):
        lines[:] = []
        try:
            fail()
            self.fail("Call should have thrown an exception")
        except AssertionError:
            pass
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine(None, "fail", "Start", ArgListNoArgs),
            self.__getLine(None, "fail", "Error", ExceptoionMsg)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedFunction_withArgs_failedCall(self):
        lines[:] = []
        try:
            fail(1, 2)
            self.fail("Call should have thrown an exception")
        except AssertionError:
            pass
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine(None, "fail", "Start", ArgListWithArgs),
            self.__getLine(None, "fail", "Error", ExceptoionMsg)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedMethod_successfulCall(self):
        lines[:] = []
        ToObserve().success()
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine("ToObserve", "success", "Start", ArgListNoArgs),
            self.__getLine("ToObserve", "success", "Done", ResultNone)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedMethod_withArgs_successfulCall(self):
        lines[:] = []
        ToObserve().success(1, 2)
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine("ToObserve", "success", "Start", ArgListWithArgs),
            self.__getLine("ToObserve", "success", "Done", ResultOne)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedMethod_failedCall(self):
        lines[:] = []
        try:
            ToObserve().fail()
            self.fail("Call should have thrown an exception")
        except AssertionError:
            pass
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine("ToObserve", "fail", "Start", ArgListNoArgs),
            self.__getLine("ToObserve", "fail", "Error", ExceptoionMsg)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def test_observedMethod_withArgs_failedCall(self):
        lines[:] = []
        try:
            ToObserve().fail(1, 2)
            self.fail("Call should have thrown an exception")
        except AssertionError:
            pass
        self.__waitForLogLines(2)
        expectedLines = sorted([
            self.__getLine("ToObserve", "fail", "Start", ArgListWithArgs),
            self.__getLine("ToObserve", "fail", "Error", ExceptoionMsg)
        ])
        actualLines = sorted(lines)
        self.assertEqual(actualLines[0], expectedLines[0])
        self.assertEqual(actualLines[1], expectedLines[1])

    def __waitForLogLines(self, count):
        while(len(lines) < count):
            time.sleep(0.005)
        self.assertEqual(len(lines), count,
                         "Expecred exactly %s lines to be observed, but was %s\nLines:\n%s" % (
            str(count), len(lines), lines))

############################################


@obs.observedMethod
def success(*args):
    if(len(args) > 0):
        return args[0]


@obs.observedMethod
def fail(*args):
    raise AssertionError("testing the error")


class ToObserve:

    @obs.observedMethod
    def success(self, *args):
        if(len(args) > 0):
            return args[0]

    @obs.observedMethod
    def fail(self, *args):
        raise AssertionError("testing the error")



from __future__ import absolute_import, print_function

import json
import importlib
import inspect

from pipecash import walletWrapper
from pipecash import agentWrapper
from pipecash import validationUtils as validate

from pipecash.observedMethod import observedMethodInstance as obs


class Scenario():

    @obs.observedMethod
    def __init__(self, path):
        with open(path, "r") as read_file:
            scenarioJson = json.load(read_file)
        self.name = scenarioJson["name"]
        self.description = scenarioJson["description"]
        self.author = scenarioJson["author"]

        for i in ["wallets", "agents", "wallet_links", "event_links", "control_links"]:
            validate.dictMember(scenarioJson, i, list)

        self.walletsJson = scenarioJson["wallets"]
        self.agentsJson = scenarioJson["agents"]
        self.wallet_links = scenarioJson["wallet_links"]
        self.event_links = scenarioJson["event_links"]
        self.control_links = scenarioJson["control_links"]

    def prepareToStart(self, secrets):
        self.secrets = secrets

        self.wallets = [walletWrapper.WalletWrapper(self.__parseWallet(w), w, secrets) for w in self.walletsJson]
        self.agents = [agentWrapper.AgentWrapper(self.__parseAgent(a), a, secrets) for a in self.agentsJson]

        for wLink in self.wallet_links:
            validate.objectType(wLink, dict, "Wallet Link")
            validate.dictMember(wLink, "source", int)
            validate.dictMember(wLink, "target", int)
            w = self.wallets[wLink["source"]]
            a = self.agents[wLink["target"]]
            a.setWallet(w)

        for eLink in self.event_links:
            validate.objectType(eLink, dict, "Event Link")
            validate.dictMember(eLink, "source", int)
            validate.dictMember(eLink, "target", int)
            source = self.agents[eLink["source"]]
            receiver = self.agents[eLink["target"]]
            receiver.receiveEventsFrom(source)

        for cLink in self.control_links:
            validate.objectType(cLink, dict, "Control Link")
            validate.dictMember(cLink, "source", int)
            validate.dictMember(cLink, "target", int)
            source = self.agents[cLink["source"]]
            receiver = self.agents[cLink["target"]]
            receiver.setControllerAgent(source)
        
    @obs.observedMethod
    def getNeededSecrets(self):

        wallets = [self.__parseWallet(w) for w in self.walletsJson]
        agents = [self.__parseAgent(a) for a in self.agentsJson]

        namesOfSecrets = []
        for i in wallets + agents:
            if hasattr(i, "uses_secret_variables"):
                namesOfSecrets = namesOfSecrets + i.uses_secret_variables

        namesOfSecrets = list(set(namesOfSecrets)) # unique names only
        return namesOfSecrets

    @obs.observedMethod
    def start(self):
        for w in self.wallets[::-1]:
            w.start()
        for a in self.agents[::-1]:
            a.start()

    def __parseWallet(self, walletJson):
        validate.dictMember(walletJson, "module", str)
        validate.dictMember(walletJson, "type", str)
        walletType = self.importClass(
            walletJson["module"], walletJson["type"])
        walletObject = walletType()

        return walletObject

    def __parseAgent(self, agentJson):
        validate.dictMember(agentJson, "module", str)
        validate.dictMember(agentJson, "type", str)
        agentType = self.importClass(
            agentJson["module"], agentJson["type"])
        agentObject = agentType()

        return agentObject

    @obs.observedMethod
    def importClass(self, moduleToImport, classToImport):
        module = importlib.import_module(moduleToImport)
        classes = inspect.getmembers(module, inspect.isclass)
        for cl in classes:
            if cl[0] == classToImport:
                return cl[1]

        raise Exception("Module %s doesn't contain class %s" % (
            moduleToImport, classToImport))



from __future__ import absolute_import, print_function

import copy

from pipecash import logWrapper


class OptionTemplateSolver:

    def solveOptions(self, optionDict, templateGlobals):
        result = {}
        templateGlobals = copy.deepcopy(templateGlobals)

        for o in optionDict.keys():
            solution = self.solve(optionDict[o], templateGlobals)
            result[o] = solution
        return result

    def solve(self, rawValue, templateGlobals):
        valueString = str(rawValue).strip()
        isMatch = valueString.startswith("{{") and valueString.endswith("}}") and len(valueString) > 4

        if not isMatch:
            return rawValue

        expression = valueString[2:-2]
        try:
            evaluation = eval(expression, templateGlobals)
            return evaluation
        except Exception as ex:
            templateGlobalsNoBuiltins = { i: templateGlobals[i] for i in templateGlobals if i[0] != "_" }
            logWrapper.loggerInstance.warning(
                "Failed to evaluate expression '%s'\n\twith data %s\n\tException: '%s'" % (
                    expression, str(templateGlobalsNoBuiltins), str(ex)))
            return str(ex)


templateSolverInstance = OptionTemplateSolver()



import unittest
import time

from pipecash import logWrapper
from pipecash.pipeObserver import Observer

from tests import utils
from tests.logMock import LogMock

logMock = LogMock(logWrapper.loggerInstance)


class ObserverTest(unittest.TestCase):

    def setUp(self):
        self.observerInstance = Observer()

    def test_observer_listen(self):
        varName = "test_observer_listen_result"
        self.observerInstance.listen("Tt", "Tt", "Tt",
                                     lambda a, n, s, o, d: setattr(self, varName, d))
        data = 42
        self.observerInstance.trigger("Tt", "Tt", "Tt", self, data)
        time.sleep(0.005)
        result = getattr(self, varName)
        key = "('Tt', 'Tt', 'Tt')"
        self.assertEqual(len(self.observerInstance._Observer__listeners), 1)
        self.assertTrue(key in self.observerInstance._Observer__listeners)
        self.assertEqual(
            len(self.observerInstance._Observer__listeners[key]), 1)
        self.assertEqual(result, data)

    def test_observer_listen_pattern(self):
        varName = "test_observer_listen_pattern_result"
        self.observerInstance.listen("Tt", None, None,
                                     lambda a, n, s, o, d: setattr(self, varName, d))
        data = 42
        self.observerInstance.trigger("Tt", "Tt", "Tt", self, data)
        time.sleep(0.005)
        result = getattr(self, varName)
        key = "('Tt', None, None)"
        self.assertEqual(len(self.observerInstance._Observer__listeners), 1)
        self.assertTrue(key in self.observerInstance._Observer__listeners)
        self.assertEqual(
            len(self.observerInstance._Observer__listeners[key]), 1)
        self.assertEqual(result, data)

    def test_observer_patterns(self):

        fullPattern = ["Tt", "Tt", "Tt"]
        allPatterns = self.observerInstance._Observer__getAllMatchingKeys(
            *fullPattern)
        expectedPatternsAll = [
            '(None, None, None)', "('Tt', 'Tt', 'Tt')",
            "('Tt', 'Tt', None)", "(None, 'Tt', 'Tt')", "('Tt', None, 'Tt')",
            "(None, None, 'Tt')", "('Tt', None, None)", "(None, 'Tt', None)", ]
        self.assertListEqual(sorted(allPatterns), sorted(expectedPatternsAll))

        nPattern = [None, "Tt", "Tt"]
        nPatterns = self.observerInstance._Observer__getAllMatchingKeys(
            *nPattern)
        expectedPatternsN = [
            '(None, None, None)', "(None, 'Tt', 'Tt')",
            "(None, None, 'Tt')", "(None, 'Tt', None)", ]
        self.assertListEqual(sorted(nPatterns), sorted(expectedPatternsN))

        limitedPattern = [None, None, "Tt"]
        limitedPatterns = self.observerInstance._Observer__getAllMatchingKeys(
            *limitedPattern)
        expectedPatternsLimited = [
            '(None, None, None)',
            "(None, None, 'Tt')", ]
        self.assertListEqual(sorted(limitedPatterns),
                             sorted(expectedPatternsLimited))

    def __wait_and_set_data(self, varName, val, secsToSleep):
        time.sleep(secsToSleep)
        setattr(self, varName, val)

    def test_observer_listenMultythreading(self):
        variables = {}

        def getMockFunc(name):
            def __listener(_a, _n, _s, _o, data):
                print("--> Writing Data to " + name + "\n")
                time.sleep(0.1)
                variables[name] = data
            return __listener

        self.observerInstance._Observer__listeners = {}
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v0"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v1"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v2"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v3"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v4"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v5"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v6"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v7"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v8"))
        self.observerInstance.listen(None, "Tt", "Tt", getMockFunc("v9"))

        key = "(None, 'Tt', 'Tt')"
        self.assertEqual(len(self.observerInstance._Observer__listeners), 1)
        self.assertTrue(key in self.observerInstance._Observer__listeners)
        self.assertEqual(len(self.observerInstance._Observer__listeners[key]), 10)

        with logMock:
            data = 42
            self.observerInstance.trigger("Tt", "Tt", "Tt", self, data)
            self.observerInstance.trigger("Tt", "Tt", "Tt", self, data)
            time.sleep(0.2)

        self.assertEqual(logMock.logs, [])

        names = ['v' + str(i) for i in range(10)]

        self.assertEqual(len(variables), 10, "Variables: " + str(variables))
        self.assertListEqual(sorted(variables.keys()), names)
        self.assertEqual(list(variables.values()), [data]*10)

    def test_observer_callbackFail(self):

        self.observerInstance.listen("Tt", "Tt", "Tt",
                                     lambda a, n, s, o, d: utils.raiseErr(NotImplementedError(
                                         "Example of a NotImplementedError message")))
        with logMock:
            self.observerInstance.trigger("Tt", "Tt", "Tt", self, None)
            time.sleep(0.005)

        expectedError = "ObserverCallbackError - ['Tt', 'Tt', 'Tt']"
        expectedError += "\nData: None "
        expectedError += "\nError: Example of a NotImplementedError message"

        self.assertEqual(len(logMock.logs), 1)
        self.assertEqual(logMock.logs[0], expectedError)



from __future__ import absolute_import, print_function

from datetime import datetime
import logging


def __createLogger(logName):
    logger = logging.getLogger(logName)

    format = '%(asctime)s: [%(threadName)s] [%(levelname)s] %(message)s'

    c_handler = logging.StreamHandler()
    formatter = logging.Formatter(format)
    c_handler.setFormatter(formatter)
    logger.addHandler(c_handler)

    return logger


def observerPrint(area, name, state, sender, data):
    dataMaxLen = 120
    dataStr = str(data)[:dataMaxLen] + \
        ("..." if len(data) > dataMaxLen else "")
    noneStr = "<None>"
    areaStr = noneStr if area is None else str(area)
    nameStr = noneStr if name is None else str(name)
    stateStr = noneStr if state is None else str(state)
    logStr = "%s.%s [%s] with data: \t\t%s" % (
        areaStr, nameStr, stateStr, dataStr)
    if state != "Error":
        loggerInstance.debug(logStr)
    else:
        loggerInstance.error(logStr)


loggerInstance = __createLogger("PipeCash")
agentLoggerInstance = __createLogger("PipeCash Agent")
walletLoggerInstance = __createLogger("PipeCash Wallet")



import unittest
import time


from pipecash.agentWrapper import AgentWrapper
from pipecash import logWrapper
from pipecash import pipeObserver

import tests.agent_test.mockAgents as mockAgents
from tests.logMock import LogMock

logMock = LogMock(logWrapper.loggerInstance)


class AgentPropagateTest(unittest.TestCase):

    def test_eventPropagation(self):
        sender1 = mockAgents.EventCreator_MockAgent()
        sender2 = mockAgents.EventCreator_MockAgent()
        target = mockAgents.EventReceiver_MockAgent()

        senderWrap1 = AgentWrapper(sender1, {
            "name": "sender1",
            "options": {
                "name": "sender1",
                "event": {"data": 3.14}
            }
        }, {})

        senderWrap2 = AgentWrapper(sender2, {
            "name": "sender2",
            "options": {
                "name": "sender2",
                "event": {"extraData": 42}
            },
            "propagate_origin_event": True
        }, {})

        targetWrap = AgentWrapper(target, {
            "name": "target",
            "options": {
                "name": "target",
                "eventData": "{{data}}"
            }
        }, {})

        senderWrap2.receiveEventsFrom(senderWrap1)
        targetWrap.receiveEventsFrom(senderWrap2)

        senderWrap1.start()
        senderWrap2.start()
        targetWrap.start()

        with logMock:
            senderWrap1._AgentWrapper__runCheck()
            time.sleep(0.005)

        expectedSender1Methods = [
            'validate_options', 'check_dependencies_missing', 'start', 'check'
        ]
        expectedSender2Methods = [
            'validate_options', 'check_dependencies_missing', 'start', 'receive'
        ]
        expectedTargetMethods = [
            'validate_options', 'check_dependencies_missing', 'start', 'receive'
        ]

        self.assertListEqual(sender1.calledMethods, expectedSender1Methods,
                             "Checking called methods of sender1...")
        self.assertListEqual(sender2.calledMethods, expectedSender2Methods,
                             "Checking called methods of sender2...")
        self.assertListEqual(target.calledMethods, expectedTargetMethods,
                             "Checking called methods of target...")
        self.assertListEqual(target.receivedData, [3.14])
        self.assertListEqual(logMock.logs, [
            "Running [Check] on EventCreator_MockAgent",
            "Running [Receive] on EventCreator_MockAgent",
            "Running [Receive] on EventReceiver_MockAgent"
        ])
        self.assertEqual(sender2.events, [str({'data': 3.14})])
        self.assertEqual(target.events, [str({'data': 3.14, 'extraData': 42})])



import unittest
import traceback

from pipecash.walletWrapper import WalletWrapper
from pipecash.secretsManager import SecretsManager

from tests import utils


def TestWalletFactory(membersToInclude):

    class WalletMock():
        pass

    w = WalletMock()

    membersDefaults = {
        "checkBalance": utils.methodMock(w, lambda: None, "checkBalance"),
        "check_dependencies_missing": utils.methodMock(w, lambda: None, "check_dependencies_missing"),
        "default_options": {},
        "description": "",
        "secrets": {},
        "getLatestTransactions": utils.methodMock(w, lambda: None, "getLatestTransactions"),
        "getReceiveAddress": utils.methodMock(w, lambda: None, "getReceiveAddress"),
        "start": utils.methodMock(w, lambda: None, "start"),
        "options": {},
        "send": utils.methodMock(w, lambda: None, "send"),
        "uses_secret_variables": lambda: True,
        "validate_options": utils.methodMock(w, lambda: None, "validate_options"),
    }

    unknownMembers = [i for i in membersToInclude if i not in membersDefaults]
    if len(unknownMembers) > 0:
        raise KeyError("Unknown wallet field : " + str(unknownMembers))

    for m in membersDefaults:
        if(m in membersToInclude):
            specified = membersToInclude[m]
            val = specified if specified is not None else membersDefaults[m]
            setattr(w, m, val)
    return w


class WalletValidationTest(unittest.TestCase):

    def __shouldPassValidation(self, members, config, secrets):
        wallet = TestWalletFactory(members)
        msg = "\n--------\nDetails:\n> With Members:\n%s\n> Config\n%s\n> Env\n%s" % (
            repr(members), repr(config), repr(secrets))
        try:
            walletWrapper = WalletWrapper(
                wallet, config, SecretsManager(secrets))
            return walletWrapper
        except Exception as ex:
            tb = traceback.format_exc()
            self.fail("Expected wallet validation to pass, but it failed :\n%s%s\n%s" % (
                str(ex), msg, tb)
            )

    def __shouldFailValidation(self, members, config, secrets):
        wallet = TestWalletFactory(members)
        msg = "\n--------\nDetails:\n> With Members:\n%s\n> Config\n%s\n> Env\n%s" % (
            repr(members), repr(config), repr(secrets))

        return utils.expectError(
            lambda: WalletWrapper(wallet, config, SecretsManager(secrets)),
            "Wallet validation was expected to fail but didn't : " + msg)

    def test_walletWrapper_validation_minimalWallet(self):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        self.__shouldPassValidation(members, config, secrets)

    def test_walletWrapper_validation_without_init_shouldFail(self):
        self.walletWrapper_validation_withoutMember_shouldFail(
            "start", "should have member : start")

    def test_walletWrapper_validation_without_checkBalance_shouldFail(self):
        self.walletWrapper_validation_withoutMember_shouldFail(
            "checkBalance", "should have member : checkBalance")

    def test_walletWrapper_validation_without_send_shouldFail(self):
        self.walletWrapper_validation_withoutMember_shouldFail(
            "send", "should have member : send")

    def test_walletWrapper_validation_without_getReceiveAddress_shouldFail(self):
        self.walletWrapper_validation_withoutMember_shouldFail(
            "getReceiveAddress", "should have member : getReceiveAddress")

    def test_walletWrapper_validation_without_getLatestTransactions_shouldFail(self):
        self.walletWrapper_validation_withoutMember_shouldFail(
            "getLatestTransactions", "should have member : getLatestTransactions")

    def walletWrapper_validation_withoutMember_shouldFail(self, exclude, errMsg):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        del members[exclude]

        ex, tb = self.__shouldFailValidation(members, config, secrets)

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         str(ex) + " did not match " + errMsg + "\n\n Occured in: \n" + tb)

    def test_walletWrapper_validation_minimalWithDefaultOptions(self):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "options": None, "default_options": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_walletWrapper_validation_withConfigOptions(self):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "options": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
            "options": {}
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_walletWrapper_validation_withConfigOptions_andValidateOptions(self):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "options": None, "validate_options": None,
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
            "options": {}
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

        called = wrapper.wallet.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "validate_options")

    def test_walletWrapper_validation_withConfigOptions_andValidateOptions(self):

        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "options": None, "validate_options": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
            "options": {}
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

        called = wrapper.wallet.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "validate_options")

    def test_walletWrapper_validation_checkDependenciesSuccess(self):
        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "check_dependencies_missing": None
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

        called = wrapper.wallet.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "check_dependencies_missing")

    def test_walletWrapper_validation_checkDependenciesFail(self):
        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "check_dependencies_missing": utils.funcToMethod(
                       lambda: utils.raiseErr(Exception("ErrMsg")))
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        ex, tb = self.__shouldFailValidation(members, config, secrets)

        self.assertEqual(str(ex),
                         "Agent 'Wallet Name' has missing dependencies: ErrMsg")

    def test_walletWrapper_validation_withEnv(self):
        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "secrets": {},
                   "uses_secret_variables": ["EXTRA_ENV_VALUE"]
                   }
        secrets = {
            "EXTRA_ENV_VALUE": "my secret value"
        }
        config = {
            "name": "Wallet Name",
        }

        wrapper = self.__shouldPassValidation(members, config, secrets)

        self.assertEqual(len(wrapper.wallet.secrets.keys()), 1)
        self.assertEqual(list(wrapper.wallet.secrets.keys())
                         [0], "EXTRA_ENV_VALUE")
        self.assertEqual(
            wrapper.wallet.secrets["EXTRA_ENV_VALUE"], "my secret value")

    def test_walletWrapper_validation_withEnv_MissingEnv(self):
        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "secrets": {},
                   "uses_secret_variables": ["EXTRA_ENV_VALUE"]
                   }
        secrets = {}
        config = {
            "name": "Wallet Name",
        }

        ex, tb = self.__shouldFailValidation(members, config, secrets)

        self.assertEqual(
            str(ex), "\"Not all keys were found : ['EXTRA_ENV_VALUE']\"")

    def test_walletWrapper_validation_withEnv_MissingWalletEnvField(self):
        members = {"start": None,
                   "checkBalance": None, "send": None,
                   "getReceiveAddress": None,
                   "getLatestTransactions": None,
                   "uses_secret_variables": ["EXTRA_ENV_VALUE"]
                   }
        secrets = {
            "EXTRA_ENV_VALUE": "my secret value"
        }
        config = {
            "name": "Wallet Name",
        }

        ex, tb = self.__shouldFailValidation(members, config, secrets)

        expectedMessageEnd = "should have member : secrets"
        self.assertEqual(
            str(ex)[-len(expectedMessageEnd):], expectedMessageEnd)



from __future__ import absolute_import, print_function

"""Console script for pipecash."""
import sys
import click
import logging

from pipecash import runner

pathType = click.Path(exists=True, file_okay=True, dir_okay=False, writable=False,
                      readable=True, resolve_path=False, allow_dash=False, path_type=None)

logLevels = ["NOTSET", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"]
logLevelValues = [
    logging.NOTSET, logging.DEBUG, logging.INFO,
    logging.WARN, logging.ERROR, logging.FATAL]

logLevelChoiceType = click.Choice(logLevels, case_sensitive=True)


@click.command()
@click.version_option("0.1.0.3")
@click.option('-s', '--scenarioPath', type=pathType, required=True, help='Path to the scenario JSON file.')
@click.option('--secretsPath', type=pathType, help='Path to the screts JSON file.')
@click.option('--createSecretsFile', is_flag=True, help="create (or print) a secrets file with blank fields and exit.")
@click.option('--logLevel', default="DEBUG", type=logLevelChoiceType, help="log level for the main log")
@click.option('--agentLogLevel', default="DEBUG", type=logLevelChoiceType, help="log level for the agent log")
@click.option('--walletLogLevel', default="DEBUG", type=logLevelChoiceType, help="log level for the wallet log")
def main(scenariopath, secretspath, createsecretsfile, loglevel, agentloglevel, walletloglevel):
    """PipeCash is an automation framework. Learn more at http://pipe.cash/"""

    if(createsecretsfile):
        runner.createSecretsFile(scenariopath,secretspath)
        return

    runner.Run(scenariopath, secretspath,
                 logLevelValues[logLevels.index(loglevel)],
                 logLevelValues[logLevels.index(agentloglevel)],
                 logLevelValues[logLevels.index(walletloglevel)]
                 )

    return 0





class TestAgent:

    def start(self, log):
        self.calledMethods.append(self.start.__name__)
        self.log = log
        pass

    def __init__(self):
        self.calledMethods = []
        self.events = []

        self.options = {}
        self.wallet = None
        self.description = ""
        self.event_description = {}
        self.default_options = {'foo': "default_value"}
        self.uses_wallet = True
        self.uses_secret_variables = []
        self.secrets = {}

    def validate_options(self):
        self.calledMethods.append(self.validate_options.__name__)

    def check(self, create_event):
        self.calledMethods.append(self.check.__name__)

    def receive(self, event, create_event):
        self.events.append(str(event))
        self.calledMethods.append(self.receive.__name__)

    def check_dependencies_missing(self):
        self.calledMethods.append(self.check_dependencies_missing.__name__)


class TestWallet:

    def start(self, log):
        self.calledMethods.append(self.start.__name__)
        self.log = log

    def __init__(self):
        self.calledMethods = []

        self.options = {}
        self.description = ""
        self.default_options = {'foo': "default_value"}
        self.uses_secret_variables = []
        self.secrets = {}

    def validate_options(self):
        self.calledMethods.append(self.validate_options.__name__)

    def check_dependencies_missing(self):
        self.calledMethods.append(self.check_dependencies_missing.__name__)

    def checkBalance(self):
        self.calledMethods.append(self.checkBalance.__name__)

    def send(self, amount, address):
        self.calledMethods.append(self.send.__name__)

    def getReceiveAddress(self):
        self.calledMethods.append(self.getReceiveAddress.__name__)

    def getLatestTransactions(self, numOfTransactions=10, transactionsToSkip=0):
        self.calledMethods.append(self.getLatestTransactions.__name__)



from __future__ import absolute_import, print_function


def objectMember(validatedObject, memberName, expectedType=None, expectedValue=None):

    if not hasattr(validatedObject, memberName):
        message = "%s should have member : %s" % (validatedObject, memberName)
        raise TypeError(message)
    if expectedType is not None:
        member = getattr(validatedObject, memberName)
        memType = type(member)
        if memType.__name__ == "unicode":
            memType = str
        if memType.__name__ == "instance":
            memType = member.__class__
        message = "Member %s.%s should be of type <%s>, but was <%s>" % (
            type(validatedObject), memberName, expectedType.__name__, memType.__name__)
        if memType != expectedType:
            raise TypeError(message)
    if expectedValue is not None:
        if getattr(validatedObject, memberName) != expectedValue:
            message = "Value of %s.%s should be : %s" % (
                type(validatedObject), memberName, expectedValue)
            raise ValueError(message)


def dictMember(validatedDictionary, memberName, expectedType=None, expectedValue=None):
    if memberName not in validatedDictionary:
        message = "%s should have member : %s" % (dict, memberName)
        raise TypeError(message)
    if expectedType is not None:
        memType = type(validatedDictionary[memberName])
        if memType.__name__ == "unicode":
            memType = str
        if memType.__name__ == "instance":
            memType = validatedDictionary[memberName].__class__
        message = "Member %s.%s should be of type <%s> but was <%s>" % (
            memType, memberName, expectedType.__name__, memType.__name__)
        if memType != expectedType:
            raise TypeError(message)
    if expectedValue is not None:
        if validatedDictionary[memberName] != expectedValue:
            message = "Value of %s.%s should be : %s" % (
                dict, memberName, expectedValue)
            raise ValueError(message)


def objectType(validatedObject, expectedType, whatIsThisObject_str="Object"):
    objType = type(validatedObject)
    if objType.__name__ == "unicode":
        objType = str
    if objType.__name__ == "instance":
        objType = validatedObject.__class__

    message = "%s <%s> should be of type <%s> but was <%s>" % (
        whatIsThisObject_str, validatedObject, expectedType.__name__, objType.__name__)

    if not issubclass(objType, expectedType):
        raise TypeError(message)



from __future__ import absolute_import, print_function


class ObservedMethod:
    observerInstance = None

    def observedMethod(self, functionOrMethod):

        def wrapper(*args, **kwargs):

            if(self.observerInstance == None):
                return functionOrMethod(*args, **kwargs)

            name = functionOrMethod.__name__

            className = None
            argNames = functionOrMethod.__code__.co_varnames
            if(argNames != None
                    and len(argNames) > 0
                    and argNames[0] == 'self'):
                selfArg = args[0]
                className = selfArg.__class__.__name__

            argsToLog = args
            if(className is not None):
                argsToLog = argsToLog[1:]

            result = None
            try:
                self.observerInstance.trigger(className, name, "Start", functionOrMethod, {
                                              'args': argsToLog, 'kwargs': kwargs})
                result = functionOrMethod(*args, **kwargs)
                self.observerInstance.trigger(
                    className, name, "Done", functionOrMethod, {'result': result})
            except Exception as error:
                self.observerInstance.trigger(
                    className, name, "Error", functionOrMethod, {'error': error})
                raise
            return result

        return wrapper


observedMethodInstance = ObservedMethod()



import unittest

import traceback
from types import MethodType, FunctionType

from pipecash.agentWrapper import AgentWrapper
from pipecash.walletWrapper import WalletWrapper
from pipecash.secretsManager import SecretsManager

import tests.utils as utils
import tests.scenario_test.mocks as mocks


def TestAgentFactory(membersToInclude):

    class AgentMock():
        calledMethods = []

    a = AgentMock()

    membersDefaults = {
        "start": utils.methodMock(a, lambda: None, "start"),
        "options": {},
        "wallet": mocks.TestWallet(),
        "description": "A few words about the agent",
        "event_description": {},
        "default_options": {},
        "default_schedule": "every_1s",
        "uses_wallet": False,
        "uses_secret_variables": [],
        "secrets": {},
        "validate_options": utils.methodMock(a, lambda: None, "validate_options"),
        "check": utils.methodMock(a, lambda: None, "check"),
        "receive": utils.methodMock(a, lambda: None, "receive"),
        "is_unavailable": utils.methodMock(a, lambda: None, "is_unavailable"),
        "check_dependencies_missing": utils.methodMock(a, lambda: None, "check_dependencies_missing"),
    }

    unknownMembers = [i for i in membersToInclude if i not in membersDefaults]
    if len(unknownMembers) > 0:
        raise KeyError("Unknown wallet field : " + str(unknownMembers))

    for m in membersDefaults:
        if(m in membersToInclude):
            specified = membersToInclude[m]
            val = specified if specified is not None else membersDefaults[m]
            setattr(a, m, val)

    return a


class AgentValidationTest(unittest.TestCase):

    def __shouldPassValidation(self, members, config, secrets):
        agent = TestAgentFactory(members)
        msg = "\n--------\nDetails:\n> With Members:\n%s\n> Config\n%s\n> Env\n%s" % (
            repr(members), repr(config), repr(secrets))
        try:
            agentWrapper = AgentWrapper(agent, config, SecretsManager(secrets))
            return agentWrapper
        except Exception as ex:
            tb = traceback.format_exc()
            self.fail("Expected agent validation to pass, but it failed :\n%s%s\n%s" % (
                str(ex), msg, tb)
            )

    def __shouldFailValidation(self, members, config, secrets):
        agent = TestAgentFactory(members)
        msg = "\n--------\nDetails:\n> With Members:\n%s\n> Config\n%s\n> Env\n%s" % (
            repr(members), repr(config), repr(secrets))

        return utils.expectError(
            lambda: AgentWrapper(agent, config, SecretsManager(secrets)),
            "Agent validation was expected to fail but didn't : " + msg)

    #   #############################
    #   #### CORRECT INPUT TESTS ####
    #   #############################

    def test_agentWrapper_validation_shouldPass_minimal(self):
        members = {"start": None, "description": None, }
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_agentWrapper_validation_shouldPass_default_schedule(self):
        members = {"start": None, "description": None,
                   "default_schedule": None, "check": None}
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_agentWrapper_validation_shouldPass_scheduleInConfig(self):
        members = {"start": None, "description": None,
                   "default_schedule": None, "check": None}
        config = {"name": "Agent Name", "schedule": "every_1s"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_agentWrapper_validation_shouldPass_check(self):
        members = {"start": None, "description": None,
                   "check": None}
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        wrapper.setControllerAgent(wrapper)

    def test_agentWrapper_validation_shouldPass_uses_secret_variables(self):
        members = {"start": None, "description": None,
                   "uses_secret_variables": None}
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

    def test_agentWrapper_validation_shouldPass_uses_secret_variables_TEST_VAR(self):
        members = {"start": None, "description": None,
                   "uses_secret_variables": ["TEST_VAR"]}
        config = {"name": "Agent Name"}
        secrets = {"TEST_VAR": "Test"}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        self.assertEqual(len(wrapper.agent.secrets), 1, wrapper.agent.secrets)
        self.assertEqual(wrapper.agent.secrets["TEST_VAR"], "Test")

    def test_agentWrapper_validation_shouldPass_default_options(self):
        members = {"start": None, "description": None,
                   "options": None, "default_options": {"foo": "bar"}}
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        self.assertEqual(wrapper.agent.options["foo"], "bar")

    def test_agentWrapper_validation_shouldPass_validate_options(self):
        members = {"start": None, "description": None,
                   "options": None, "default_options": None, "validate_options": None}
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        called = wrapper.agent.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "validate_options")

    def test_agentWrapper_validation_shouldPass_validate_options_fromConfig(self):
        members = {"start": None, "description": None,
                   "options": None, "validate_options": None}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        called = wrapper.agent.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "validate_options")

    def test_agentWrapper_validation_shouldPass_uses_wallet(self):
        members = {"start": None, "description": None,
                   "wallet": None, "uses_wallet": True}
        config = {"name": "Agent Name", "options": {}}
        secrets = {"WALLET_KEY": "TestKey"}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        wallet = WalletWrapper(mocks.TestWallet(), {
                               "name": "W1"}, SecretsManager(secrets))
        wrapper.setWallet(wallet)
        self.assertEqual(wrapper.walletWrapper, wallet)
        self.assertEqual(wrapper.walletWrapper.wallet, wallet.wallet)

    def test_agentWrapper_validation_shouldPass_receive(self):
        members = {"start": None, "description": None,
                   "receive": None}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        wrapper.receiveEventsFrom(wrapper)

    def test_agentWrapper_validation_shouldPass_check_dependencies_missing(self):
        members = {"start": None, "description": None,
                   "check_dependencies_missing": None}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)
        called = wrapper.agent.calledMethods
        self.assertEqual(len(called), 1)
        self.assertEqual(called[0], "check_dependencies_missing")

    #   #############################
    #   ### TESTS EXPECTED ERRORS ###
    #   #############################

    def test_agentWrapper_validation_lessThanMinimal_shouldFail_noInit(self):
        members = {"description": None, }
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)

        errMsg = "should have member : start"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_lessThanMinimal_shouldFail_NoDescription(self):
        members = {"start": None, }
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)

        errMsg = "should have member : description"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_minimal_wrongInitType(self):
        members = {"start": lambda: None, "description": None, }
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)

        errMsg = ".start should be of type <%s>, but was <%s>" % (
            MethodType.__name__, FunctionType.__name__)

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_minimal_wrongDescriptionType(self):
        members = {"start": None, "description": 42, }
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)

        errMsg = ".description should be of type <str>, but was <int>"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_defaultSchedule_noCheck(self):
        members = {"start": None, "description": None,
                   "default_schedule": None}
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "should have member : check"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_scheduleInConfig_noCheck(self):
        members = {"start": None, "description": None, }
        config = {"name": "Agent Name", "schedule": "every_1s"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "should have member : check"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_controlled_noCheck(self):
        members = {"start": None, "description": None, }
        config = {"name": "Agent Name"}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

        errMsg = "should have member : check"

        ex, tb = utils.expectError(
            lambda: wrapper.setControllerAgent(wrapper),
            "Setting Controller Agent should have failed")
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_defaultScheduleIsWrongType(self):
        members = {"start": None, "description": None,
                   "default_schedule": 42, "check": None}
        config = {"name": "Agent Name"}
        secrets = {}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = ".default_schedule should be of type <str>, but was <int>"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_CheckIsWrongType(self):
        members = {"start": None, "description": None,
                   "default_schedule": None, "check": 42}
        config = {"name": "Agent Name"}
        secrets = {}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = ".check should be of type <%s>, but was <int>" % MethodType.__name__
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_scheduleInConfig_wrongType(self):
        members = {"start": None, "description": None,
                   "default_schedule": None, "check": None}
        config = {"name": "Agent Name", "schedule": 1}
        secrets = {}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = ".schedule should be of type <str> but was <int>"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_secrets_WrongVarName(self):
        members = {"start": None, "description": None,
                   "secrets": None, "uses_secret_variables": ["TEST_VAR"]}
        config = {"name": "Agent Name"}
        secrets = {"TEST_VAR_WrongVarName": "Test"}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "\"Not all keys were found : ['TEST_VAR']\""
        self.assertEqual(str(ex), errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_secrets_usedVarsNotAList(self):
        members = {"start": None, "description": None,
                   "uses_secret_variables": {"TEST_VAR": "42"}}
        config = {"name": "Agent Name"}
        secrets = {"TEST_VAR": "Test"}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "should be of type <list>, but was <dict>"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_secrets_usedVarNamesNotStrings(self):
        members = {"start": None, "description": None,
                   "uses_secret_variables": [42]}
        config = {"name": "Agent Name"}
        secrets = {"42": "Test"}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "Agent Secrets Variable Request <42> should be of type <str> but was <int>"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_validate_options_wrongType(self):
        members = {"start": None, "description": None,
                   "validate_options": 42, "default_options": None, "options": None}
        config = {"name": "Agent Name"}
        secrets = {}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = ".validate_options should be of type <%s>, but was <int>" % MethodType.__name__
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_noOptionSource(self):
        members = {"start": None, "description": None,
                   "options": None}
        config = {"name": "Agent Name"}
        secrets = {}

        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "should have member : default_options"
        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_setWallet_withNonWallet(self):
        members = {"start": None, "description": None,
                   "wallet": None, "uses_wallet": True}
        config = {"name": "Agent Name", "options": {}}
        secrets = {"WALLET_KEY": "TestKey"}
        wallet = 42
        wrapper = self.__shouldPassValidation(members, config, secrets)
        errMsg = "should be of type <WalletWrapper> but was <int>"

        ex, tb = utils.expectError(lambda: wrapper.setWallet(wallet),
                                   "Setting wallet should have failed")

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_setWallet_withWalletInsteadOfWalletWrapper(self):
        members = {"start": None, "description": None,
                   "wallet": None, "uses_wallet": True}
        config = {"name": "Agent Name", "options": {}}
        secrets = {"WALLET_KEY": "TestKey"}
        wallet = mocks.TestWallet()
        wrapper = self.__shouldPassValidation(members, config, secrets)
        errMsg = "should be of type <WalletWrapper> but was <TestWallet>"

        ex, tb = utils.expectError(lambda: wrapper.setWallet(wallet),
                                   "Setting wallet should have failed")

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_checkDependenciesMissing_raises(self):
        members = {"start": None, "description": None,
                   "check_dependencies_missing": utils.funcToMethod(
                       lambda: utils.raiseErr(Exception("Dependancy Missing")))}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "Agent 'Agent Name' has missing dependencies: Dependancy Missing"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_validateOptions_raises(self):
        members = {"start": None, "description": None,
                   "options": None, "default_options": None,
                   "validate_options": utils.funcToMethod(
                       lambda: utils.raiseErr(Exception("Invalid Options")))}
        config = {"name": "Agent Name"}
        secrets = {}
        ex, tb = self.__shouldFailValidation(members, config, secrets)
        errMsg = "validate_options of AgentMock failed : Invalid Options"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_receive_wrongType(self):
        members = {"start": None, "description": None,
                   "receive": 42}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

        ex, tb = utils.expectError(lambda: wrapper.receiveEventsFrom(wrapper),
                                   "receiveEventsFrom Wrapper should fail if 'receive' is not a method")

        errMsg = ".receive should be of type <%s>, but was <int>" % MethodType.__name__

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_receive_wrongSenderType(self):
        members = {"start": None, "description": None,
                   "receive": None}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

        ex, tb = utils.expectError(lambda: wrapper.receiveEventsFrom(42),
                                   "receiveEventsFrom Wrapper should fail if 'receive' is not a method")

        errMsg = "Agent Event Sender <42> should be of type <AgentWrapper> but was <int>"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))

    def test_agentWrapper_validation_shouldFail_receive_senderIsNone(self):
        members = {"start": None, "description": None,
                   "receive": None}
        config = {"name": "Agent Name", "options": {}}
        secrets = {}
        wrapper = self.__shouldPassValidation(members, config, secrets)

        ex, tb = utils.expectError(lambda: wrapper.receiveEventsFrom(None),
                                   "receiveEventsFrom Wrapper should fail if 'receive' is not a method")

        errMsg = "Agent Event Sender <None> should be of type <AgentWrapper> but was <NoneType>"

        self.assertEqual(str(ex)[-len(errMsg):], errMsg,
                         "Actual Error did not match expected: \nActual  :%s\nExpected:%s\n----\n Occured in: \n%s" % (str(ex), errMsg, tb))



import time


class MockAgent:

    def start(self, log):
        self.calledMethods.append(self.start.__name__)
        self.log = log
        pass

    def __init__(self):
        self.calledMethods = []
        self.events = []
        self.description = "Test Agent " + self.__class__.__name__
        self.options = {}
        self.default_options = {}

    def validate_options(self):
        self.calledMethods.append(self.validate_options.__name__)

    def check(self, create_event):
        self.calledMethods.append(self.check.__name__)

    def receive(self, event, create_event):
        self.events.append(str(event))
        self.calledMethods.append(self.receive.__name__)

    def check_dependencies_missing(self):
        self.calledMethods.append(self.check_dependencies_missing.__name__)


class ThreadingTest_MockAgent(MockAgent):
    counter = 0

    def start(self, log):
        MockAgent.start(self, log)
        self.counter = 0

    def check(self, create_event):
        c = self.counter
        time.sleep(0.001)
        self.counter = c + 1
        return MockAgent.check(self, create_event)

    def receive(self, event, create_event):
        c = self.counter
        time.sleep(0.002)
        self.counter = c + 1
        return MockAgent.receive(self, event, create_event)


class EventReceiver_MockAgent(MockAgent):

    def __init__(self):
        MockAgent.__init__(self)
        self.receivedData = []
        self.default_options = {
            "eventData": "{{data}}"
        }

    def check(self, create_event):
        self.receivedData.append(self.options["eventData"])
        return MockAgent.check(self, create_event)

    def receive(self, event, create_event):
        self.receivedData.append(self.options["eventData"])
        return MockAgent.receive(self, event, create_event)


class EventCreator_MockAgent(MockAgent):

    def __init__(self):
        MockAgent.__init__(self)
        self.default_options = {
            "event": {"data": 42}
        }
        self.event_description = { "data": 42 }

    def check(self, create_event):
        create_event(self.options["event"])
        return MockAgent.check(self, create_event)

    def receive(self, event, create_event):
        create_event(self.options["event"])
        return MockAgent.receive(self, event, create_event)



from __future__ import absolute_import, print_function

import schedule as pythonScheduler
import threading

from types import MethodType
import time

from pipecash import validationUtils as validate
from pipecash.observedMethod import observedMethodInstance as obs


class Scheduler:
    __secondsPerLetter = {"s": 1, "m": 60, "h": 60*60, "d": 60*60*24}

    __stop = False

    def everyScheduleToSeconds(self, sch):
        every, t = str(sch).split("_")
        if(every != "every" or len(t) < 2):
            raise LookupError(
                "Can't parse schedule '%s' as seconds" % str(sch))
        num = int(t[0:-1])
        letter = t[-1:]
        return num * self.__secondsPerLetter[letter]

    def atScheduleToTime(self, sch):
        if(sch == "midnight"):
            return "00:00"
        if(sch == "noon"):
            return "12:00"
        hour = int(sch[0:-2])
        isPM = sch[-2:] == "pm"
        return "%i:00" % (hour + (12 if isPM else 0))

    @obs.observedMethod
    def __init__(self):
        self.__knownSchedules_every = [
            "every_1s", "every_2s", "every_5s", "every_10s", "every_30s",
            "every_1m", "every_2m", "every_5m", "every_10m", "every_30m",
            "every_1h", "every_2h", "every_5h", "every_12h",
            "every_1d", "every_2d", "every_7d", "every_30d"
        ]
        self.__knownSchedules_at = [
            "midnight", "1am", "2am", "3am", "4am", "5am",
            "6am", "7am", "8am", "9am", "10am", "11am",
            "noon", "1pm", "2pm", "3pm", "4pm", "5pm",
            "6pm", "7pm", "8pm", "9pm", "10pm", "11pm",
        ]

        self.__knownSchedules_other = ["never"]

        self.__knownSchedules_every_s = [self.everyScheduleToSeconds(
            i) for i in self.__knownSchedules_every]
        self.__knownSchedules_at_time = [
            self.atScheduleToTime(i) for i in self.__knownSchedules_at]
        self.__knownSchedules = self.__knownSchedules_every + \
            self.__knownSchedules_at + self.__knownSchedules_other
        self.__scheduledTasks = dict((k, []) for k in self.__knownSchedules)

    def isKnownSchedule(self, sch):
        return sch in self.knownSchedules()

    def knownSchedules(self):
        return list(self.__knownSchedules)

    @obs.observedMethod
    def registerTask(self, sch, task):
        if(not self.isKnownSchedule(sch)):
            raise IndexError("Unknown schedule '%s'" % str(sch))
        validate.objectMember(task, "__call__")
        self.__scheduledTasks[sch].append(task)

    @obs.observedMethod
    def start(self):
        self.__stop = False
        for sch in self.__knownSchedules_every:
            if any(self.__scheduledTasks[sch]):
                index = self.__knownSchedules_every.index(sch)
                s = self.__knownSchedules_every_s[index]
                self.__everyNSeconds_do(s, sch)
        for sch in self.__knownSchedules_at:
            if any(self.__scheduledTasks[sch]):
                index = self.__knownSchedules_at.index(sch)
                time = self.__knownSchedules_at_time[index]
                self.__eachDayAtTime_do(time, sch)

        thread = threading.Thread(target=self.__schedulerLoop)
        thread.setName("SCH")
        thread.start()

    def __schedulerLoop(self):
        while not self.__stop:
            pythonScheduler.run_pending()
            time.sleep(1)

    def __everyNSeconds_do(self, seconds, schedule):
        pythonScheduler.every(seconds).seconds.do(lambda: self.__job(schedule))

    def __eachDayAtTime_do(self, time, schedule):
        pythonScheduler.every().day.at(time).do(lambda: self.__job(schedule))

    @obs.observedMethod
    def stop(self):
        self.__stop = True

    def __job(self, triggeredSchedule):
        if self.__stop is True:
            return
        tasks = self.__scheduledTasks[triggeredSchedule]
        threads = []
        for t in tasks:
            thread = threading.Thread(target=t)
            thread.setName("SCH(%s)" % triggeredSchedule)
            thread.start()
            threads.append(thread)
        for t in threads:
            t.join()


schedulerInstance = Scheduler()



import unittest
import threading

from pipecash.agentWrapper import AgentWrapper
import tests.agent_test.mockAgents as mockAgents


class AgentThreadingTest(unittest.TestCase):

    def test_agentMustHave_Only1Thread(self):
        agent = mockAgents.ThreadingTest_MockAgent()
        wrapper = AgentWrapper(agent, {"name": "threading-test"}, {})
        wrapper.start()

        calledMethods = ['validate_options',
                         'check_dependencies_missing', 'start']

        self.assertListEqual(agent.calledMethods, calledMethods)
        agent.calledMethods = []

        def do100times(action): return [action() for i in range(100)]

        def run100ChecksLambda(): return do100times(
            lambda: wrapper._AgentWrapper__runCheck())

        def run100ReceivesLambda(): return do100times(
            lambda: wrapper._AgentWrapper__receiveEvent(
                "area", "name", "state", self, {}))

        threads = []
        for i in range(5):
            threads.append(threading.Thread(target=run100ChecksLambda))
            threads.append(threading.Thread(target=run100ReceivesLambda))

        for t in threads:
            t.start()

        for t in threads:
            t.join()

        expectedCount = 5*2*100
        self.assertEqual(agent.counter, expectedCount)
        self.assertEqual(len(agent.calledMethods), expectedCount)



import logging


class HandlerMock(logging.Handler):
    def __init__(self, logCollection):
        self.logCollection = logCollection
        return super(HandlerMock, self).__init__(1)

    def emit(self, record):
        self.logCollection.append(record.msg)


class LogMock():
    logs = []

    def __init__(self, logger):
        self.logger = logger
        self.originalLevel = logger.level
        self.originalHandlers = list(logger.handlers)

    def __enter__(self):
        self.logs = []
        for h in list(self.logger.handlers):
            self.logger.removeHandler(h)
        self.logger.addHandler(HandlerMock(self.logs))
        self.logger.setLevel(1)

    def __exit__(self, type, value, tb):
        for h in list(self.logger.handlers):
            self.logger.removeHandler(h)
        for h in self.originalHandlers:
            self.logger.addHandler(h)
        self.logger.setLevel(self.originalLevel)



# -*- coding: utf-8 -*-

"""Top-level package for pipecash."""

__author__ = """Aleksandar Nikolaev Dinkov"""
__email__ = 'alexander.n.dinkov@gmail.com'
__version__ = '0.1.0.3'

from pipecash import agentWrapper
from pipecash import cli
from pipecash import logWrapper
from pipecash import observedMethod
from pipecash import optionTemplateSolver
from pipecash import pipeObserver
from pipecash import pipeScheduler
from pipecash import runner
from pipecash import scenario
from pipecash import secretsManager
from pipecash import validationUtils
from pipecash import walletWrapper


from __future__ import absolute_import, print_function

from pipecash import scenario
from pipecash import secretsManager
from pipecash import pipeObserver
from pipecash import observedMethod
from pipecash import logWrapper
from pipecash import pipeScheduler

import time
import json

def Run(scenariopath, secretspath, loglevel=1, agentloglevel=1, walletloglevel=1):

    if(secretspath is not None):
        secrets = secretsManager.SecretsManager.loadFromFile(secretspath)
    else:
        secrets = secretsManager.SecretsManager({})

    observedMethod.observedMethodInstance.observerInstance = pipeObserver.observerInstance
    pipeObserver.observerInstance.listen(
        None, None, None, logWrapper.observerPrint)

    logWrapper.loggerInstance.setLevel(loglevel)
    logWrapper.agentLoggerInstance.setLevel(agentloglevel)
    logWrapper.walletLoggerInstance.setLevel(walletloglevel)

    logWrapper.loggerInstance.info("Starting PipeCash...")

    sc = scenario.Scenario(scenariopath)
    sc.prepareToStart(secrets)
    sc.start()

    pipeScheduler.schedulerInstance.start()
    while(True):
        time.sleep(1)

def createSecretsFile(scenariopath, secretspath):
    sc = scenario.Scenario(scenariopath)
    namesOfSecrets = sc.getNeededSecrets()
    secretsDict = dict(zip(namesOfSecrets, [ "" ] * len(namesOfSecrets)))
    secretsJson = json.dumps(secretsDict, indent=2)

    if secretspath == None:
        print(secretsJson)
    else:
        with open(secretspath, "w") as writeFile:
            writeFile.write(secretsJson)


import unittest

from pipecash import validationUtils as vld


class MyObj(object):
    foo = None


class ValidationUtilsTest(unittest.TestCase):

    def __verify_A_withFunc_F_andArgs_B_shouldPass(self, A, F, *B):
        try:
            F(A, *B)
        except Exception as ex:
            self.fail("Failed Attempt to verify %s with %s(%s)\n\tError: %s" % (
                repr(A), F.__name__, repr(B), ex))

    def __verify_A_withFunc_F_args_B_shouldFail(self, A, F, *B):
        try:
            F(A, *B)
        except Exception as ex:
            print("\nExpected error was cought: " + str(ex))
            return ex

        self.fail("Expected failure when verifying %s with %s(%s)" % (
            repr(A), F.__name__, repr(B)))

    # TESTS FOR SUCCESSFUL VALIDATION

    def test_verify_objectType_42_is_int(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            42, vld.objectType, int)

    def test_verify_objectType_42_is_typeof42(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            42, vld.objectType, type(42))

    def test_verify_objectType_42_is_int_withDescription(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            42, vld.objectType, int, "The Answer")

    def test_verify_objectType_Foo_is_str(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            "foo", vld.objectType, str)

    def test_verify_objectType_obj_is_MyObj(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            MyObj(), vld.objectType, MyObj)

    def test_verify_objectType_obj_is_parentClass(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            MyObj(), vld.objectType, object)

    def test_verify_objectType_None_is_NoneType(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            None, vld.objectType, type(None))

    def test_verify_dictMember_dict_Foo_is_str_Bar(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            {"foo": "bar"}, vld.dictMember, "foo", str, "bar")

    def test_verify_dictMember_dict_Foo_is_int_42(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            {"foo": 42}, vld.dictMember, "foo", int, 42)

    def test_verify_dictMember_dict_Foo_is_None(self):
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            {"foo": None}, vld.dictMember, "foo", type(None), None)

    def test_verify_objectMember_dict_Foo_is_str_Bar(self):
        obj = MyObj()
        obj.foo = "bar"
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            obj, vld.objectMember, "foo", str, "bar")

    def test_verify_objectMember_dict_Foo_is_int_42(self):
        obj = MyObj()
        obj.foo = 42
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            obj, vld.objectMember, "foo", int, 42)

    def test_verify_objectMember_dict_Foo_is_None(self):
        obj = MyObj()
        obj.foo = None
        self.__verify_A_withFunc_F_andArgs_B_shouldPass(
            obj, vld.objectMember, "foo", type(None), None)

    # TESTS FOR FAILED VALIDATION

    def test_verifyFail_objectType_42_is_stringInstance(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            42, vld.objectType, "")

    def test_verifyFail_objectType_42_is_noneInstance(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            42, vld.objectType, None)

    def test_verifyFail_objectType_42_is_str_withDescription(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            42, vld.objectType, str, "The Answer")

    def test_verify_objectType_str_Foo_is_int(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            "foo", vld.objectType, int)

    def test_verify_objectType_obj_is_noParamPassed(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            MyObj(), vld.objectType)

    def test_verify_objectType_obj_is_intInstance(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            MyObj(), vld.objectType, 42)

    def test_verify_objectType_NoneType_is_ofType_NoneType(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            type(None), vld.objectType, type(None))

    def test_verify_dictMember_dict_Foo_is_int_Bar(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            {"foo": "bar"}, vld.dictMember, "foo", int, "bar")

    def test_verify_dictMember_dict_Foo_is_str_wrongValue(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            {"foo": "bar"}, vld.dictMember, "foo", str, "wrong value")

    def test_verify_dictMember_dict_Foo_is_int_strValue(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            {"foo": 42}, vld.dictMember, "foo", int, str(42))

    def test_verify_dictMember_dict_Foo_is_noTypeArgPassed(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            {"foo": None}, vld.dictMember, "foo", "bar", "baz")

    def test_verify_dictMember_noParams(self):
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            {"foo": None}, vld.dictMember)

    def test_verify_objectMember_dict_Foo_is_int_Bar(self):
        obj = MyObj()
        obj.foo = "bar"
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            obj, vld.objectMember, "foo", int, "bar")

    def test_verify_objectMember_dict_Foo_is_str_wrongValue(self):
        obj = MyObj()
        obj.foo = "bar"
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            obj, vld.objectMember, "foo", str, "wrong value")

    def test_verify_objectMember_dict_Foo_is_int_strValue(self):
        obj = MyObj()
        obj.foo = 42
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            obj, vld.objectMember, "foo", int, str(42))

    def test_verify_objectMember_dict_Foo_is_noTypeArgPassed(self):
        obj = MyObj()
        obj.foo = "foo"
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            obj, vld.objectMember, "foo", "bar", "baz")

    def test_verify_objectMember_noParams(self):
        obj = MyObj()
        obj.foo = "bar"
        ex = self.__verify_A_withFunc_F_args_B_shouldFail(
            obj, vld.objectMember)



from __future__ import absolute_import, print_function

import threading

from pipecash import logWrapper


class Observer:
    __listeners = {}

    def __init__(self):
        self.__listeners = {}

    def listen(self, area, name, state, callback):
        '''Subscribe to the observed event.
        When the event occurs, the callback will be called.
        'None' value is considered a wildcard and will match
        all patterns in that position. If 3 None values are
        passed for area, name and state, the callback will
        be called on all events.

        The callback will receive 5 arguments:
        > area, name, state, sender, data'''

        key = str((area, name, state))
        if key not in self.__listeners:
            self.__listeners[key] = []
        self.__listeners[key].append(callback)

    def trigger(self, area, name, state, sender, data):
        '''
        Trigger the observed event.
        Notifies the observer that something has happened.
            area - where did it happen
            name - name of the event
            state - state of the event
            sender - the object asociated with the event
            data - details about the event
        '''

        keys = self.__getAllMatchingKeys(area, name, state)
        for k in keys:
            if k in self.__listeners:
                callbacks = self.__listeners[k]
                threads = []
                for c in callbacks:
                    thread = threading.Thread(target=lambda: self.__call(
                        c, area, name, state, sender, data))
                    thread.setName("OBS")
                    thread.start()
                    threads.append(thread)
                for t in threads:
                    t.join()

    def __call(self, callback, area, name, state, sender, data):
        try:
            callback(area, name, state, sender, data)
        except Exception as ex:
            logWrapper.loggerInstance.error(
                "ObserverCallbackError - %s\nData: %s \nError: %s" % (
                    repr([area, name, state]), repr(data), ex))

    def __getAllMatchingKeys(self, a, b, c):
        _ = None
        patterns = [
            (a, b, c), (_, b, c), (a, _, c), (a, b, _),
            (a, _, _), (_, b, _), (_, _, c), (_, _, _),
        ]
        keys = list(set([str(p) for p in patterns]))
        return keys


observerInstance = Observer()



import unittest
import sys

from pipecash.optionTemplateSolver import templateSolverInstance
from pipecash import logWrapper

from tests import utils
from tests.logMock import LogMock

logMock = LogMock(logWrapper.loggerInstance)


class OptionTemplateSolverTest(unittest.TestCase):

    def __test_output(self, line):
        self.logs.append(str(line))

    variables = {
        "a": 10,
        "b": 5,
        "s": "s",
    }

    def test_emptyBrackets_shouldProduce_emptyBrackets(self):
        self.__solving_A_shouldProduce_B('{{}}', '{{}}')

    def test_normalString_shouldProduce_normalString(self):
        self.__solving_A_shouldProduce_B('normal string', 'normal string')

    def test_42_shouldProduce_42(self):
        self.__solving_A_shouldProduce_B(42, 42)

    def test_True_shouldProduce_True(self):
        self.__solving_A_shouldProduce_B(True, True)

    def test_VariableA_shouldProduce_ValueOfA(self):
        self.__solving_A_shouldProduce_B('{{a}}', 10)

    def test_StrOnVarA_shouldProduce_StrA(self):
        self.__solving_A_shouldProduce_B('{{str(a)}}', '10')

    def test_APlusB_shouldProduce_Sum(self):
        self.__solving_A_shouldProduce_B('{{a+b}}', 15)

    def test_VariableInText_shouldProduce_SameText(self):
        self.__solving_A_shouldProduce_B(
            'in the {{a}} middle', 'in the {{a}} middle')

    def test_VariableInWord_shouldProduce_SameText(self):
        self.__solving_A_shouldProduce_B('also mi{{a}}ddle', 'also mi{{a}}ddle')

    def test_StrOnAPlusS_shouldProduce_10s(self):
        self.__solving_A_shouldProduce_B('{{str(a)+s}}', '10s')

    def test_IncorrectFormatting_shouldProduce_NoChange(self):
        self.__solving_A_shouldProduce_B('{ {a}}', '{ {a}}')

    def test_VarS_shouldProduce_ValS(self):
        self.__solving_A_shouldProduce_B('{{s}}', 's')

    def test_SPlusS_shouldProduce_SS(self):
        self.__solving_A_shouldProduce_B('{{s+s}}', 'ss')

    def test_RangeB_shouldProduce_ZeroToFour(self):
        self.__solving_A_shouldProduce_B(
            '{{list(range(b))}}', [0, 1, 2, 3, 4])

    def test_LenOfStr_shouldProduce_2(self):
        self.__solving_A_shouldProduce_B('{{len(s+s)}}', 2)

    def test_StringConstantsCombined_shouldProduce_ABC(self):
        self.__solving_A_shouldProduce_B("{{'ab' + 'c'}}", 'abc')

    def __solving_A_shouldProduce_B(self, a, b):
        results = templateSolverInstance.solveOptions(
            {"data": a}, self.variables)
        resultData = results["data"]
        assertMessage = "Expected resolving of %s\nto produce: %s\nbut it was: %s" % (
            repr(a), repr(b), resultData)
        self.assertEqual(resultData, b, assertMessage)

    def test_solving_UndefinedVariable_shlouldLogException(self):
        self.__solving_A_shlouldLogException("{{c}}")

    def test_solving_DevisionByTwo_shlouldLogException(self):
        self.__solving_A_shlouldLogException("{{1/0}}")

    def test_solving_SumStringAndInt_shlouldLogException(self):
        self.__solving_A_shlouldLogException("{{s+a}}")

    def test_TwoTagsNextToEachOther_shlouldLogException(self):
        self.__solving_A_shlouldLogException("{{1}}{{2}}")

    def test_TwoTagsFarFromEachOther_shlouldLogException(self):
        self.__solving_A_shlouldLogException("{{1}} Same Text {{a}}")


    def __solving_A_shlouldLogException(self, expression):
        msgStart = "Failed to evaluate expression"
        msgExMarker = "Exception: "

        with logMock:
            results = templateSolverInstance.solveOptions(
                {"data": expression}, self.variables)
            resultData = results["data"]


        self.assertEqual(len(logMock.logs), 1, "Expected exactly 1 warning message")

        expected = "".join([
            msgStart,
            " '", 
            expression[2:-2], 
            "'\n\twith data ", 
            str(self.variables), 
            "\n\tException: '",
            resultData,
            "'"
        ])
        self.assertEqual(logMock.logs[0], expected, "Message was not as expected.")

        exceptionIndex = logMock.logs[0].index(msgExMarker)+len(msgExMarker)
        exceptionMsg = logMock.logs[0][exceptionIndex:]




import unittest

from pipecash.secretsManager import SecretsManager

secretsPath = "tests/general_test/test_secretsManager.json"


class SecretsManagerTest(unittest.TestCase):

    def setUp(self):
        pass

    def test_secrets_parsing(self):
        secrets = SecretsManager.loadFromFile(secretsPath)
        names = ['v' + str(i) for i in range(8)]
        self.assertListEqual(secrets.findMissingKeys(*names), [])
        self.assertFalse(secrets.hasKey("v8"))
        values = secrets.get(*names)
        for n in names:
            self.assertEqual(values[n], 42)

    def test_secrets_getSubset(self):
        secrets = SecretsManager.loadFromFile(secretsPath)
        names = ['v' + str(i) for i in range(4)]
        self.assertListEqual(secrets.findMissingKeys(*names), [])
        values = secrets.get(*names)
        for n in names:
            self.assertEqual(values[n], 42)
        self.assertEqual(len(values.keys()), 4)

    def test_secrets_noSuchKey(self):
        secrets = SecretsManager.loadFromFile(secretsPath)
        self.assertListEqual(secrets.findMissingKeys(
            "v1", "noKey", "v2"), ["noKey"])
        with self.assertRaises(KeyError):
            secrets.get("v1", "noKey", "v2")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=6.0', 'schedule==0.5.0' ]

setup_requirements = [ ]

test_requirements = [ 'schedule==0.5.0' ]

setup(
    author="Aleksandar Nikolaev Dinkov",
    author_email='alexander.n.dinkov@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="A framework for easy automation of Bitcoin related tasks. PipeCash is flexible due to it's plugin system and configuration options. Read more at http://pipe.cash/",
    entry_points={
        'console_scripts': [
            'pipecash=pipecash.cli:main',
        ],
    },
    install_requires=requirements,
    license="OPEN BLOCKCHAIN-SPECIFIC LICENSE",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pipecash',
    name='pipecash',
    packages=find_packages(include=['pipecash']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Pipe-Cash/pipecash',
    version='0.1.0.3',
    zip_safe=False,
)



from __future__ import absolute_import, print_function

from types import MethodType

from pipecash import logWrapper
from pipecash import validationUtils as validate

from pipecash.observedMethod import observedMethodInstance as obs


class WalletWrapper:

    @obs.observedMethod
    def __init__(self, _wallet, _configuration, _secrets):
        self.wallet = _wallet
        self.configuration = _configuration

        self.type = type(self.wallet).__name__
        self.name = self.configuration["name"]
        validate.objectType(self.name, str, "Wallet Name")

        validate.objectMember(self.wallet, "start", MethodType)
        validate.objectMember(self.wallet, "checkBalance", MethodType)
        validate.objectMember(self.wallet, "send", MethodType)
        validate.objectMember(self.wallet, "getReceiveAddress", MethodType)
        validate.objectMember(self.wallet, "getLatestTransactions", MethodType)

        self.__initOptions()
        self.__initSecrets(_secrets)
        self.__checkDependencies()

    @obs.observedMethod
    def start(self):
        self.wallet.start(logWrapper.walletLoggerInstance)

    def __initOptions(self):
        if not hasattr(self.wallet, "options"):
            return
        validate.objectMember(self.wallet, "options", dict)
        if "options" in self.configuration:
            validate.dictMember(self.configuration, "options", dict)
            self.wallet.options = self.configuration["options"]
        else:
            validate.objectMember(self.wallet, "default_options", dict)
            self.wallet.options = self.wallet.default_options

        if(hasattr(self.wallet, "validate_options")):
            validate.objectMember(self.wallet, "validate_options", MethodType)
            try:
                self.wallet.validate_options()
            except Exception as ex:
                raise Exception("validate_options of %s failed" %
                                self.type, ex)

    def __initSecrets(self, _secrets):
        if hasattr(self.wallet, "uses_secret_variables"):
            validate.objectMember(self.wallet, "secrets")
            validate.objectMember(self.wallet, "uses_secret_variables", list)
            if len(self.wallet.uses_secret_variables) > 0:
                for i in self.wallet.uses_secret_variables:
                    validate.objectType(
                        i, str, "Wallet Secrets Variable Request")
                secrets = _secrets.get(*self.wallet.uses_secret_variables)
                self.secrets = secrets
                self.wallet.secrets = secrets

    def __checkDependencies(self):
        if hasattr(self.wallet, "check_dependencies_missing"):
            validate.objectMember(
                self.wallet, "check_dependencies_missing", MethodType)
            try:
                self.wallet.check_dependencies_missing()
            except Exception as ex:
                raise EnvironmentError(
                    "Agent '%s' has missing dependencies: %s" % (self.name, ex))

    def __mergeDictionaries(self, dict1, dict2):
        result = dict1.copy()   # start with dict1's keys and values
        # modifies result with dict2's keys and values & returns None
        result.update(dict2)
        return result



from __future__ import absolute_import, print_function

from types import MethodType
from uuid import uuid4
from threading import Lock

from pipecash import validationUtils as validate
from pipecash import walletWrapper
from pipecash import optionTemplateSolver
from pipecash import pipeObserver
from pipecash import pipeScheduler
from pipecash import logWrapper

from pipecash.observedMethod import observedMethodInstance as obs


class AgentWrapper:

    @obs.observedMethod
    def __init__(self, _agent, _configuration, _secrets):

        self.__lock = Lock()
        self.__id = hash(id(self))

        self.walletWrapper = None
        self.options = {}
        self.propagate_origin_event = False
        self.configuration = _configuration
        self.agent = _agent

        # Use "__class__" instead of "type()" for Python2 compatablity
        self.type = self.agent.__class__.__name__
        validate.dictMember(self.configuration, "name", str)
        self.name = self.configuration["name"]

        validate.objectMember(self.agent, "start", MethodType)
        validate.objectMember(self.agent, "description", str)

        self.__initOptions()
        self.__initSchedule()
        self.__initSecrets(_secrets)
        self.__checkDependencies()
        self.__initEventConfiguration()
        self.__initConditions()
        self.__initExtendEvent()

    @obs.observedMethod
    def start(self):
        self.agent.start(logWrapper.agentLoggerInstance)

    @obs.observedMethod
    def setWallet(self, _walletWrapper):
        validate.objectMember(self.agent, "wallet")
        validate.objectMember(self.agent, "uses_wallet", bool, True)
        self.walletWrapper = _walletWrapper
        validate.objectType(self.walletWrapper,
                            walletWrapper.WalletWrapper, "Agent Wallet")
        self.agent.wallet = self.walletWrapper.wallet

    @obs.observedMethod
    def receiveEventsFrom(self, _sender):
        validate.objectMember(self.agent, "receive", MethodType)
        validate.objectType(_sender, AgentWrapper, "Agent Event Sender")
        pipeObserver.observerInstance.listen(
            None, "Event", _sender.__id, self.__receiveEvent)

    @obs.observedMethod
    def setControllerAgent(self, _sender):
        validate.objectMember(self.agent, "check", MethodType)
        validate.objectType(_sender, AgentWrapper, "Agent Controller")
        pipeObserver.observerInstance.listen(
            None, "Event", _sender.__id, self.__receiveControlEvent)

    def isWorking(self):
        return self.__lock.locked()

    def __createEvent(self, eventDict):
        if type(eventDict) is not dict:
            logWrapper.loggerInstance.warning(
                "Agent '%s' created an event of a wrong type (%s). Ignoring." % (
                    self.name, type(eventDict)))
            return

        if not self.__checkCondition("event_condition"):
            return

        if hasattr(self, "extend_event") and self.extend_event is not None:
            eventDict = self.__mergeDictionaries(self.extend_event, eventDict)

        pipeObserver.observerInstance.trigger(
            self.name, "Event", self.__id, self, eventDict)

    def __getEventCreator(self, currentEvent=None):
        if self.propagate_origin_event is not True:
            return self.__createEvent
        elif currentEvent is None:
            return self.__createEvent
        else:
            return lambda newEvent: self.__createEvent(self.__mergeDictionaries(currentEvent, newEvent))

    def __receiveEvent(self, area, name, state, sender, eventData):
        if not self.__checkCondition("receive_condition"):
            return

        logWrapper.loggerInstance.info(
            "Running [Receive] on " + self.agent.__class__.__name__)
        create_event = self.__getEventCreator(eventData)
        self.__runAction(lambda: self.agent.receive(eventData, create_event), "Receive_Event", True, eventData)

    def __receiveControlEvent(self, area, name, state, sender, eventData):
        if not self.__checkCondition("control_condition"):
            return

        logWrapper.loggerInstance.info(
            "Running [Check] on " + self.agent.__class__.__name__)
        create_event = self.__getEventCreator(eventData)
        self.__runAction(lambda: self.agent.check(create_event), "Control_Check", True, eventData)

    def __runCheck(self, *args):
        logWrapper.loggerInstance.info(
            "Running [Check] on " + self.agent.__class__.__name__)
        create_event = self.__getEventCreator(None)
        self.__runAction(lambda: self.agent.check(create_event), "Scheduled_Check", False, {})

    def __runAction(self, action, actionName, solveOptions=False, eventData={}):
        self.__lock.acquire()

        originalOptions = self.options
        if solveOptions:
            solvedOptions = optionTemplateSolver.templateSolverInstance.solveOptions(
                originalOptions, eventData)
            self.agent.options = solvedOptions

        try:
            action()
        except Exception as ex:
            pipeObserver.observerInstance.trigger(
                self.type, actionName, "Error", self, ex)
        finally:
            self.agent.options = originalOptions
            self.__lock.release()

    def __checkCondition(self, conditionName):
        result = False

        if not hasattr(self, conditionName):
            return True
        condition = getattr(self, conditionName)
        if condition is None:
            return True

        try:
            result = condition()
            if result == True:
                return True
            elif result == False:
                logWrapper.loggerInstance.debug(
                    "event_condition on %s retuned False" % (
                        self.agent.__class__.__name__))
                return False
            else:
                raise AssertionError("Evaluation result must be True or False")
        except Exception as ex:
            logWrapper.loggerInstance.error("Failed to check %s = %s.\n%s" % (
                conditionName, self.configuration[conditionName], str(ex)))
            return False

    def __initSecrets(self, _secrets):
        if "uses_secret_variables" in self.configuration:
            validate.dictMember(self.configuration,
                                "uses_secret_variables", list)
            if not hasattr(self.agent, "uses_secret_variables"):
                self.agent.uses_secret_variables = []
            else:
                validate.objectMember(
                    self.agent, "uses_secret_variables", list)
            for varName in self.configuration["uses_secret_variables"]:
                self.agent.uses_secret_variables.append(varName)

        if hasattr(self.agent, "uses_secret_variables"):
            validate.objectMember(self.agent, "uses_secret_variables", list)
            if len(self.agent.uses_secret_variables) > 0:
                for i in self.agent.uses_secret_variables:
                    validate.objectType(
                        i, str, "Agent Secrets Variable Request")
                self.secrets = _secrets.get(*self.agent.uses_secret_variables)
                self.agent.secrets = self.secrets

    def __initOptions(self):
        if not hasattr(self.agent, "options"):
            return

        if "options" in self.configuration:
            validate.dictMember(self.configuration, "options", dict)
            validate.objectMember(self.agent, "options")
            self.agent.options = self.configuration["options"]
        else:
            validate.objectMember(self.agent, "default_options", dict)
            self.agent.options = self.agent.default_options

        self.options = self.agent.options

        if(hasattr(self.agent, "validate_options")):
            validate.objectMember(self.agent, "validate_options", MethodType)
            try:
                self.agent.validate_options()
            except Exception as ex:
                raise Exception("validate_options of %s failed : %s" %
                                (self.type, str(ex)))

    def __initSchedule(self):
        if hasattr(self.agent, "default_schedule") or "schedule" in self.configuration:
            validate.objectMember(self.agent, "check", MethodType)
            if "schedule" in self.configuration:
                validate.dictMember(self.configuration, "schedule", str)
                pipeScheduler.schedulerInstance.registerTask(
                    self.configuration["schedule"],
                    self.__runCheck)
            elif hasattr(self.agent, "default_schedule"):
                validate.objectMember(self.agent, "default_schedule", str)
                pipeScheduler.schedulerInstance.registerTask(
                    self.agent.default_schedule,
                    self.__runCheck)

    def __initEventConfiguration(self):
        if "propagate_origin_event" in self.configuration:
            validate.dictMember(self.configuration,
                                "propagate_origin_event", bool)
            self.propagate_origin_event = self.configuration["propagate_origin_event"]

    def __initConditions(self):
        _secrets = None
        if hasattr(self.agent, "secrets"):
            _secrets = self.agent.secrets

        if "event_condition" in self.configuration:
            validate.dictMember(self.configuration, "event_condition", str)
            self.event_condition = self.__getConditionEvaluatorFunc(
                self.configuration["event_condition"],
                _secrets)
        if "receive_condition" in self.configuration:
            validate.objectMember(self.agent, "receive", MethodType)
            validate.dictMember(self.configuration, "receive_condition", str)
            self.receive_condition = self.__getConditionEvaluatorFunc(
                self.configuration["receive_condition"],
                _secrets)
        if "control_condition" in self.configuration:
            validate.objectMember(self.agent, "check", MethodType)
            validate.dictMember(self.configuration, "control_condition", str)
            self.control_condition = self.__getConditionEvaluatorFunc(
                self.configuration["control_condition"],
                _secrets)

    def __initExtendEvent(self):
        if "extend_event" in self.configuration:
            validate.dictMember(self.configuration, "extend_event", dict)
            self.extend_event = self.configuration["extend_event"]

    def __getConditionEvaluatorFunc(self, condition, secrets):
        addon = {}
        if secrets is not None:
            addon = {"secrets": secrets}

        def solve(globsDict): return bool(optionTemplateSolver.templateSolverInstance.solve(
            condition, self.__mergeDictionaries(globsDict, addon)))
        return solve

    def __checkDependencies(self):
        if hasattr(self.agent, "check_dependencies_missing"):
            validate.objectMember(
                self.agent, "check_dependencies_missing", MethodType)
            try:
                self.agent.check_dependencies_missing()
            except Exception as ex:
                raise EnvironmentError(
                    "Agent '%s' has missing dependencies: %s" % (self.name, str(ex)))

    def __mergeDictionaries(self, dict1, dict2):
        result = dict1.copy()   # start with dict1's keys and values
        # modifies result with dict2's keys and values & returns None
        result.update(dict2)
        return result
